from typing import NamedTuple, Sequence
from dataclasses import dataclass, field
import asyncio
from uuid import uuid4
from sqlmodel import Session
from sqlalchemy import Engine
from haskellian import either as E, dicts as D
from kv import KV
import pure_cv as vc
from pipeteer import WriteQueue
from moveread.dfy import Game, queries
from moveread.pipelines.dfy import Input
from dslog import Logger
import robust_extraction2 as re
import scoresheet_models as sm

def inputId(tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId}/{group}/{round}/{board}_{uuid4()}'

def title(game: Game, tournId: str, group: str, round: str, board: str) -> str:
  s = f'{tournId} {group}/{round}/{board} {game.white} - {game.black}'
  if game.result:
    s += f' {game.result}'
  return s

class NewInput(NamedTuple):
  input: Input
  """Pipeline input"""
  from_urls: Sequence[str]
  to_urls: Sequence[str]

def new_input(uuid: str, game: Game, *, model_name: str, model: re.ExtendedModel):
  gid = game.gameId()
  from_urls = [img.url for img in game.imgs]
  to_urls = [f'{uuid}/{i}.jpg' for i in range(len(from_urls))]
  endpoint = f'/v1/models/{gid["tournId"]}-{gid["group"]}:predict'
  task = Input(gameId=gid, model=model, model_name=model_name, imgs=to_urls, title=title(game, **gid), serving_endpoint=endpoint)
  return NewInput(input=task, from_urls=from_urls, to_urls=to_urls)

@dataclass
class Puller:
  Qpush: WriteQueue[Input]
  pipeline_blobs: KV[bytes]
  dfy_blobs: KV[bytes]
  models: sm.ModelsCache = field(default_factory=sm.ModelsCache)
  logger: Logger = field(default_factory=lambda: Logger.click().prefix('[PULLER]'))

  @E.do()
  async def jpg_copy(self, url_from: str, url_to: str):
    """Copies the image by downloading it first, encoding it as JPG, then inserting"""
    img = (await self.dfy_blobs.read(url_from)).mapl(lambda e: f'Error reading dfy blob "{url_from}": {e}').unsafe()
    jpg = vc.encode(vc.decode(img), '.jpg')
    return (await self.pipeline_blobs.insert(url_to, jpg)).mapl(lambda e: f'Error inserting blob "{url_to}": {e}').unsafe()

  async def copy_images(self, from_urls: Sequence[str], to_urls: Sequence[str]):
    tasks = [self.jpg_copy(from_, to) for from_, to in zip(from_urls, to_urls)]
    results = await asyncio.gather(*tasks)
    return E.sequence(results)

  @E.do()
  async def pull_one(self, game: Game, model_name: str):
    model = (await self.models.fetch(model_name)).mapl(lambda e: f'Error fetching model "{model_name}: {e}"').unsafe()
    id = inputId(**game.gameId())
    x = new_input(id, game, model=model, model_name=model_name)
    (await self.copy_images(x.from_urls, x.to_urls)).mapl(lambda e: f'Error copying images: {e}').unsafe()
    (await self.Qpush.push(id, x.input)).mapl(lambda e: f'Error pushing task: {e}').unsafe()


  async def pull_games(self, session: Session, games: Sequence[Game]):
    tnmt_games = D.group_by(lambda g: g.tournId, games)
    for tournId, games in tnmt_games.items():
      if not (model := queries.Select(session).model(tournId)):
        self.logger(f'No model for tournament "{tournId}"', level='WARNING')
        continue

      for game in games:
        r = await self.pull_one(game, model)
        if r.tag == 'left':
          self.logger(f'Error pulling game {game.gameId()}: {r.value}', level='ERROR')
        else:
          game.status = 'doing'
          session.add(game)
          session.commit()
          self.logger(f'Pulled game {game.gameId()}')

  async def loop(self, dfy_engine: Engine, interval: float = 300):
    while True:
      with Session(dfy_engine) as session:
        games = queries.Select(session).uploaded_games()
        if games:
          self.logger(f'Pulling {len(games)} games...')
          await self.pull_games(session, games)
        else:
          self.logger('No games to pull', level='DEBUG')
          
      await asyncio.sleep(interval)


import os
from argparse import ArgumentParser

def env(variable: str, *, default = None, required: bool = True) -> dict:
  if (value := os.getenv(variable, default)) is not None:
    return dict(default=value)
  return dict(required=required)

def main():
  parser = ArgumentParser(description='DFY Puller')
  parser.add_argument('--dfy-sql', **env('DFY_SQL'), help='SQL connection string')
  parser.add_argument('--dfy-blobs', **env('DFY_BLOBS'), help='DFY Blobs connection string')
  parser.add_argument('--pipeline', **env('PIPELINE_ENDPOINT'), help='Pipeline endpoint')
  parser.add_argument('--token', **env('PIPELINE_TOKEN'), help='Pipeline token')

  parser.add_argument('--interval', **env('PULL_INTERVAL', default=60), type=int, help='Interval between pulls')

  args = parser.parse_args()
  endpoint = args.pipeline.rstrip('/')

  from dslog import Logger
  logger = Logger.click().prefix('[PULLER]')
  logger(f'Starting puller...')
  logger(f'- Endpoint: {args.pipeline}')
  logger(f'- Interval: {args.interval}s')

  import asyncio
  from kv import KV, ClientKV
  from sqlmodel import create_engine
  from pipeteer import http
  from moveread.pipelines.dfy import DFYPipeline
  from moveread.dfy_connector import Puller

  HEADERS = { 'Authorization': f'Bearer {args.token}' }
  req = http.bound_request(headers=HEADERS)
  pipe = DFYPipeline()
  Qpush, *_ = http.clients(pipe, f'{endpoint}/queues', request=req)
  blobs = ClientKV(f'{endpoint}/images', request=req)
  online_blobs = KV.of(args.dfy_blobs)
  engine = create_engine(args.dfy_sql)

  puller = Puller(Qpush=Qpush, pipeline_blobs=blobs, dfy_blobs=online_blobs, logger=logger)
  asyncio.run(puller.loop(engine, args.interval))

if __name__ == '__main__':
  print('Executing as main')
  from dotenv import load_dotenv
  load_dotenv()
  main()


from dataclasses import dataclass, field
import asyncio
from sqlmodel import Session
from sqlalchemy import Engine
from haskellian import either as E
from kv import KV
from pipeteer import ReadQueue
from dslog import Logger
import chess_pairings as cp
from moveread.core import Core
from moveread.dfy import PGN, queries
from moveread.pipelines import dfy

@dataclass
class Pusher:
  Qout: ReadQueue[dfy.Output]
  pipeline_blobs: KV[bytes]
  dfy_blobs: KV[bytes]
  dfy_engine: Engine
  core: Core
  logger: Logger = field(default_factory=lambda: Logger.click().prefix('[PUSHER]'))

  async def push_game(self, output: dfy.Output):
    gid = output.gameId
    key = cp.stringifyId(**gid)
    e = await dfy.core.output_one(self.core, key, output, blobs=self.pipeline_blobs)
    if e.tag == 'left':
      self.logger(f'Error pushing output for game {output.gameId}:', e.value, level='ERROR')
      return False

    with Session(self.dfy_engine) as ses:
      api = queries.API(ses)
      game = api.select.game(**gid)
      if game is None:
        self.logger(f'Game not found: {gid}', level='ERROR')
        return False
      
      if game.pgn:
        ses.delete(game.pgn)
        ses.commit()

      game.pgn = PGN(moves=output.pgn, early=output.early)
      game.status = 'done'
      ses.add(game)
      ses.commit()
      self.logger(f'Pushed game: {gid}')
      return True
  
  @E.do()
  async def push_one(self):
    id, output = (await self.Qout.read()).unsafe()
    self.logger(f'Pushing "{id}"')
    r = await self.push_game(output)
    if r:
      (await self.Qout.pop(id)).unsafe()
      self.logger(f'Popped "{id}"')
      (await self.pipeline_blobs.prefix(id).clear()).unsafe()
      self.logger(f'Cleaned blobs of "{id}"')

  async def loop(self):
    while True:
      try:
        r = await self.push_one()
        if r.tag == 'left':
          self.logger('Error pushing:', r.value, level='ERROR')
          await asyncio.sleep(5)
      except Exception as e:
        self.logger('Unexpected exception:', e, level='ERROR')
        await asyncio.sleep(5)



import os
from argparse import ArgumentParser

def env(variable: str, *, default = None, required: bool = True) -> dict:
  if (value := os.getenv(variable, default)) is not None:
    return dict(default=value)
  return dict(required=required)

def main():
  parser = ArgumentParser(description='DFY Pusher')
  parser.add_argument('--dfy-sql', **env('DFY_SQL'), help='DFY SQL connection string')
  parser.add_argument('--dfy-blobs', **env('DFY_BLOBS'), help='DFY Blobs KV connection string')
  parser.add_argument('--pipeline', **env('PIPELINE_ENDPOINT'), help='Pipeline endpoint')
  parser.add_argument('--token', **env('PIPELINE_TOKEN'), help='Pipeline token')
  parser.add_argument('--core-meta', **env('CORE_META'), help='Core metadata KV connection string')
  parser.add_argument('--core-blobs', **env('CORE_BLOBS'), help='Core blobs KV connection string')


  args = parser.parse_args()
  endpoint = args.pipeline.rstrip('/')

  from dslog import Logger
  logger = Logger.click().prefix('[PUSHER]')
  logger(f'Starting pusher...')
  logger(f'- Endpoint: {args.pipeline}')

  import asyncio
  from kv import KV, ClientKV
  from sqlmodel import create_engine
  from pipeteer import http
  from moveread.core import Core
  from moveread.pipelines.dfy import DFYPipeline
  from moveread.dfy_connector import Pusher

  HEADERS = { 'Authorization': f'Bearer {args.token}' }
  req = http.bound_request(headers=HEADERS)
  pipe = DFYPipeline()
  _, Qout, *_ = http.clients(pipe, f'{endpoint}/queues', request=req)
  blobs = ClientKV(f'{endpoint}/images', request=req)
  online_blobs = KV.of(args.dfy_blobs)
  engine = create_engine(args.dfy_sql)
  core = Core.of(args.core_meta, args.core_blobs)

  pusher = Pusher(
    Qout=Qout, pipeline_blobs=blobs, dfy_blobs=online_blobs,
    dfy_engine=engine, core=core, logger=logger
  )
  asyncio.run(pusher.loop())

if __name__ == '__main__':
  print('Executing as main')
  from dotenv import load_dotenv
  load_dotenv()
  main()